document.addEventListener('DOMContentLoaded', function () {
  docsearch({
    container: '#docsearch',
    appId: 'CTX8NKR4GC',
    apiKey: 'eb05e80b7739efec7038d98158f39a54',
    indexName: 'getdaft',
    debug: false,
    insights: true,
  });
});
